
/*******************************************************************************
#             uvccapture: USB UVC Video Class Snapshot Software                #
#This package work with the Logitech UVC based webcams with the mjpeg feature  #
#.                                                                             #
# 	Orginally Copyright (C) 2005 2006 Laurent Pinchart &&  Michel Xhaard   #
#       Modifications Copyright (C) 2006  Gabriel A. Devenyi                   #
#       Modifications Copyright (C) 2011  Nonoo <nonoo@nonoo.hu>               #
#	Modifications Copyright (C) 2011  Reece Arnott <reece.arnott@gmail.com>#
#                                                                              #
# This program is free software; you can redistribute it and/or modify         #
# it under the terms of the GNU General Public License as published by         #
# the Free Software Foundation; either version 2 of the License, or            #
# (at your option) any later version.                                          #
#                                                                              #
# This program is distributed in the hope that it will be useful,              #
# but WITHOUT ANY WARRANTY; without even the implied warranty of               #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                #
# GNU General Public License for more details.                                 #
#                                                                              #
# You should have received a copy of the GNU General Public License            #
# along with this program; if not, write to the Free Software                  #
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA    #
#                                                                              #
*******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <jpeglib.h>
#include <time.h>
#include <linux/videodev2.h>

#include "v4l2uvc.h"

#define MAXFILENAMELENGTH 50

static const char version[] = VERSION;
int run = 1;

void
sigcatch (int sig)
{
  fprintf (stderr, "Exiting...\n");
  run = 0;
}

void
usage (void)
{
  fprintf (stderr, "uvccapture version %s\n", version);
  fprintf (stderr, "Usage is: uvccapture [options]\n");
  fprintf (stderr, "Options:\n");
  fprintf (stderr, "-v\t\tVerbose\n");
  fprintf (stderr, "-o<filename>\tOutput filename(default: snap.jpg)\n");
  fprintf (stderr, "-d<device>\tV4L2 Device(default: /dev/video0)\n");
  fprintf (stderr,
	   "-b\t\tRun in the background\n");
  fprintf (stderr,
	   "-x<width>\tImage Width(must be supported by device)\n");
  fprintf (stderr,
	   "-y<height>\tImage Height(must be supported by device)\n");
  fprintf (stderr,
	   "-c<command>\tCommand to run after each image capture(executed as <command> <output_filename>)\n");
  fprintf (stderr,
	   "-i<command>\tInitialising command to run before each image capture(executed as <command> <device>)\n");
  fprintf (stderr,
	   "-t<integer>\tTake continuous shots with <integer> seconds between them (0 for single shot)\n");
  fprintf (stderr,
	   "-q<percentage>\tJPEG Quality Compression Level\n");
  fprintf (stderr, "-r\t\tUse read instead of mmap for image capture\n");
  fprintf (stderr,
	   "-w\t\tWait for pre or post capture commands to finish before continuing\n");
  fprintf (stderr, "-m\t\tToggles capture mode to YUYV capture\n");
  fprintf (stderr,
	   "-D<integer>\tPre-delay before capture\n");
  fprintf (stderr,
	   "-P<integer>\tPost-delay after capture\n");
  fprintf (stderr,
	   "-W\t\tConstantly capture images during the pre-delay and post-delay loops\n");
  fprintf (stderr, "Camera Settings:\n");
  fprintf (stderr, "-B<integer>\tBrightness\n");
  fprintf (stderr, "-C<integer>\tContrast\n");
  fprintf (stderr, "-S<integer>\tSaturation\n");
  fprintf (stderr, "-G<integer>\tGain\n");
  exit (8);
}

static unsigned char dht_data[] =
{
	0xff, 0xc4, 0x01, 0xa2, 0x00, 0x00, 0x01, 0x05, 0x01, 0x01,
	0x01, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08,
	0x09, 0x0a, 0x0b, 0x10, 0x00, 0x02, 0x01, 0x03, 0x03, 0x02,
	0x04, 0x03, 0x05, 0x05, 0x04, 0x04, 0x00, 0x00, 0x01, 0x7d,
	0x01, 0x02, 0x03, 0x00, 0x04, 0x11, 0x05, 0x12, 0x21, 0x31,
	0x41, 0x06, 0x13, 0x51, 0x61, 0x07, 0x22, 0x71, 0x14, 0x32,
	0x81, 0x91, 0xa1, 0x08, 0x23, 0x42, 0xb1, 0xc1, 0x15, 0x52,
	0xd1, 0xf0, 0x24, 0x33, 0x62, 0x72, 0x82, 0x09, 0x0a, 0x16,
	0x17, 0x18, 0x19, 0x1a, 0x25, 0x26, 0x27, 0x28, 0x29, 0x2a,
	0x34, 0x35, 0x36, 0x37, 0x38, 0x39, 0x3a, 0x43, 0x44, 0x45,
	0x46, 0x47, 0x48, 0x49, 0x4a, 0x53, 0x54, 0x55, 0x56, 0x57,
	0x58, 0x59, 0x5a, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68, 0x69,
	0x6a, 0x73, 0x74, 0x75, 0x76, 0x77, 0x78, 0x79, 0x7a, 0x83,
	0x84, 0x85, 0x86, 0x87, 0x88, 0x89, 0x8a, 0x92, 0x93, 0x94,
	0x95, 0x96, 0x97, 0x98, 0x99, 0x9a, 0xa2, 0xa3, 0xa4, 0xa5,
	0xa6, 0xa7, 0xa8, 0xa9, 0xaa, 0xb2, 0xb3, 0xb4, 0xb5, 0xb6,
	0xb7, 0xb8, 0xb9, 0xba, 0xc2, 0xc3, 0xc4, 0xc5, 0xc6, 0xc7,
	0xc8, 0xc9, 0xca, 0xd2, 0xd3, 0xd4, 0xd5, 0xd6, 0xd7, 0xd8,
	0xd9, 0xda, 0xe1, 0xe2, 0xe3, 0xe4, 0xe5, 0xe6, 0xe7, 0xe8,
	0xe9, 0xea, 0xf1, 0xf2, 0xf3, 0xf4, 0xf5, 0xf6, 0xf7, 0xf8,
	0xf9, 0xfa, 0x01, 0x00, 0x03, 0x01, 0x01, 0x01, 0x01, 0x01,
	0x01, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09, 0x0a,
	0x0b, 0x11, 0x00, 0x02, 0x01, 0x02, 0x04, 0x04, 0x03, 0x04,
	0x07, 0x05, 0x04, 0x04, 0x00, 0x01, 0x02, 0x77, 0x00, 0x01,
	0x02, 0x03, 0x11, 0x04, 0x05, 0x21, 0x31, 0x06, 0x12, 0x41,
	0x51, 0x07, 0x61, 0x71, 0x13, 0x22, 0x32, 0x81, 0x08, 0x14,
	0x42, 0x91, 0xa1, 0xb1, 0xc1, 0x09, 0x23, 0x33, 0x52, 0xf0,
	0x15, 0x62, 0x72, 0xd1, 0x0a, 0x16, 0x24, 0x34, 0xe1, 0x25,
	0xf1, 0x17, 0x18, 0x19, 0x1a, 0x26, 0x27, 0x28, 0x29, 0x2a,
	0x35, 0x36, 0x37, 0x38, 0x39, 0x3a, 0x43, 0x44, 0x45, 0x46,
	0x47, 0x48, 0x49, 0x4a, 0x53, 0x54, 0x55, 0x56, 0x57, 0x58,
	0x59, 0x5a, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68, 0x69, 0x6a,
	0x73, 0x74, 0x75, 0x76, 0x77, 0x78, 0x79, 0x7a, 0x82, 0x83,
	0x84, 0x85, 0x86, 0x87, 0x88, 0x89, 0x8a, 0x92, 0x93, 0x94,
	0x95, 0x96, 0x97, 0x98, 0x99, 0x9a, 0xa2, 0xa3, 0xa4, 0xa5,
	0xa6, 0xa7, 0xa8, 0xa9, 0xaa, 0xb2, 0xb3, 0xb4, 0xb5, 0xb6,
	0xb7, 0xb8, 0xb9, 0xba, 0xc2, 0xc3, 0xc4, 0xc5, 0xc6, 0xc7,
	0xc8, 0xc9, 0xca, 0xd2, 0xd3, 0xd4, 0xd5, 0xd6, 0xd7, 0xd8,
	0xd9, 0xda, 0xe2, 0xe3, 0xe4, 0xe5, 0xe6, 0xe7, 0xe8, 0xe9,
	0xea, 0xf2, 0xf3, 0xf4, 0xf5, 0xf6, 0xf7, 0xf8, 0xf9, 0xfa
};

// from fswebcam: https://github.com/fsphil/fswebcam/blob/master/dec_jpeg.c
int verify_jpeg_dht(unsigned char *src, int lsrc, unsigned char **dst, int *ldst, int verbose)
{
	/* This function is based on a patch provided by Scott J. Bertin. */

	unsigned char *p, *i = NULL;

	/* By default we simply return the source image. */
	*dst = src;
	*ldst = lsrc;

	/* Scan for an existing DHT segment or the first SOS segment. */
	for(p = src + 2; p - src < lsrc - 3 && i == NULL; )
	{
		if(*(p++) != 0xFF) continue;

		if(*p == 0xD9) break;           /* JPEG_EOI */
		if(*p == 0xC4) return(0);       /* JPEG_DHT */
		if(*p == 0xDA && !i) i = p - 1; /* JPEG_SOS */

		/* Move to next segment. */
		p += (p[1] << 8) + p[2];
	}

	/* If no SOS was found, insert the DHT directly after the SOI. */
	if(i == NULL) i = src + 2;

        if (debug)
	    printf("Inserting DHT segment into JPEG frame.\n");

	*ldst = lsrc + sizeof(dht_data);
	*dst  = malloc(*ldst);
	if(!*dst)
	{
    	    if (verbose >= 1)
		fprintf(stderr,"Out of memory.\n");
		return(-1);
	}

	/* Copy the JPEG data, inserting the DHT segment. */
	memcpy((p  = *dst), src, i - src);
	memcpy((p += i - src), dht_data, sizeof(dht_data));
	memcpy((p += sizeof(dht_data)), i, lsrc - (i - src));

	return(1);
}

int
compress_yuyv_to_jpeg (struct vdIn *vd, FILE * file, int quality)
{
  struct jpeg_compress_struct cinfo;
  struct jpeg_error_mgr jerr;
  JSAMPROW row_pointer[1];
  unsigned char *line_buffer, *yuyv;
  int z;

  line_buffer = calloc (vd->width * 3, 1);
  yuyv = vd->framebuffer;

  cinfo.err = jpeg_std_error (&jerr);
  jpeg_create_compress (&cinfo);
  jpeg_stdio_dest (&cinfo, file);

  cinfo.image_width = vd->width;
  cinfo.image_height = vd->height;
  cinfo.input_components = 3;
  cinfo.in_color_space = JCS_RGB;

  jpeg_set_defaults (&cinfo);
  jpeg_set_quality (&cinfo, quality, TRUE);

  jpeg_start_compress (&cinfo, TRUE);

  z = 0;
  while (cinfo.next_scanline < cinfo.image_height) {
    int x;
    unsigned char *ptr = line_buffer;

    for (x = 0; x < vd->width; x++) {
      int r, g, b;
      int y, u, v;

      if (!z)
	y = yuyv[0] << 8;
      else
	y = yuyv[2] << 8;
      u = yuyv[1] - 128;
      v = yuyv[3] - 128;

      r = (y + (359 * v)) >> 8;
      g = (y - (88 * u) - (183 * v)) >> 8;
      b = (y + (454 * u)) >> 8;

      *(ptr++) = (r > 255) ? 255 : ((r < 0) ? 0 : r);
      *(ptr++) = (g > 255) ? 255 : ((g < 0) ? 0 : g);
      *(ptr++) = (b > 255) ? 255 : ((b < 0) ? 0 : b);

      if (z++) {
	z = 0;
	yuyv += 4;
      }
    }

    row_pointer[0] = line_buffer;
    jpeg_write_scanlines (&cinfo, row_pointer, 1);
  }

  jpeg_finish_compress (&cinfo);
  jpeg_destroy_compress (&cinfo);

  free (line_buffer);

  return (0);
}

int
spawn (char *argv[], int wait, int verbose)
{
  pid_t pid;
  int rv;

  switch (pid = fork ()) {
  case -1:
    return -1;
  case 0:
    // CHILD
    execvp (argv[0], argv);
    fprintf (stderr, "Error executing command '%s'\n", argv[0]);
    exit (1);
  default:
    // PARENT
    if (wait == 1) {
      if (verbose >= 1)
	fprintf (stderr, "Waiting for command to finish...");
      waitpid (pid, &rv, 0);
      if (verbose >= 1)
	fprintf (stderr, "\n");
    } else {
      // Clean zombies
      waitpid (-1, &rv, WNOHANG);
      rv = 0;
    }
    break;
  }

  return rv;
}

int
main (int argc, char *argv[])
{
  char *videodevice = "/dev/video0";
  char outputfile[MAXFILENAMELENGTH];
  char timedfilename[MAXFILENAMELENGTH];
  char *pre_capture_command[3];
  char *post_capture_command[3];
  int format = V4L2_PIX_FMT_MJPEG;
  int grabmethod = 1;
  int width = 320;
  int height = 240;
  int brightness = 0, contrast = 0, saturation = 0, gain = 0;
  int verbose = 0;
  int delay = 0;
  int quality = 95;
  int command_wait = 0;
  time_t ref_time;
  time_t save_time;
  struct tm timestruct;
  struct vdIn *videoIn;
  FILE *file;
  unsigned char* imageWithDHT = NULL;
  int imageWithDHTSize = 0;
  int preDelay = 0;
  int postDelay = 0;
  int backGround = 0;
  int preDelayCapture = 0;
  int cameraControlsEnabled = 0;

  (void) signal (SIGINT, sigcatch);
  (void) signal (SIGQUIT, sigcatch);
  (void) signal (SIGKILL, sigcatch);
  (void) signal (SIGTERM, sigcatch);
  (void) signal (SIGABRT, sigcatch);
  (void) signal (SIGTRAP, sigcatch);

  // set pre_capture_command to default values
  pre_capture_command[0] = NULL;
  pre_capture_command[1] = NULL;
  pre_capture_command[2] = NULL;
  // set post_capture_command to default values
  post_capture_command[0] = NULL;
  post_capture_command[1] = NULL;
  post_capture_command[2] = NULL;

  outputfile[0] = 0;

  //Options Parsing (FIXME)
  while ((argc > 1) && (argv[1][0] == '-')) {
    switch (argv[1][1]) {
    case 'v':
      verbose++;
      break;

    case 'o':
      snprintf(outputfile, MAXFILENAMELENGTH, "%s", &argv[1][2]);
      break;

    case 'b':
      backGround = 1;
      break;

    case 'd':
      videodevice = &argv[1][2];
      break;

    case 'W':
      preDelayCapture = 1;
      break;

    case 'x':
      width = atoi (&argv[1][2]);
      break;

    case 'y':
      height = atoi (&argv[1][2]);
      break;

    case 'r':
      grabmethod = 0;
      break;

    case 'm':
      format = V4L2_PIX_FMT_YUYV;
      break;

    case 't':
      delay = atoi (&argv[1][2]);
      break;

    case 'c':
      post_capture_command[0] = &argv[1][2];
      break;

    case 'i':
      pre_capture_command[0] = &argv[1][2];
      break;

    case 'w':
      command_wait = 1;
      break;

    case 'B':
      brightness = atoi (&argv[1][2]);
      cameraControlsEnabled = 1;
      break;

    case 'C':
      contrast = atoi (&argv[1][2]);
      cameraControlsEnabled = 1;
      break;

    case 'S':
      saturation = atoi (&argv[1][2]);
      cameraControlsEnabled = 1;
      break;

    case 'G':
      gain = atoi (&argv[1][2]);
      cameraControlsEnabled = 1;
      break;

    case 'q':
      quality = atoi (&argv[1][2]);
      break;

    case 'D':
      preDelay = atoi (&argv[1][2]);
      break;

    case 'P':
      postDelay = atoi (&argv[1][2]);
      break;

    case 'h':
      usage ();
      break;

    default:
      fprintf (stderr, "Unknown option %s \n", argv[1]);
      usage ();
    }
    ++argv;
    --argc;
  }

//  if ((width > 960) || (height > 720) || (quality != 95))
//    format = V4L2_PIX_FMT_YUYV;

  if (post_capture_command[0])
    post_capture_command[1] = outputfile;
  if (verbose >= 1) {
    fprintf (stderr, "Using videodevice: %s\n", videodevice);
    fprintf (stderr, "Saving images to: %s\n", outputfile);
    fprintf (stderr, "Image size: %dx%d\n", width, height);
    fprintf (stderr, "Taking snapshot every %d seconds\n", delay);
    if (grabmethod == 1)
      fprintf (stderr, "Taking images using mmap\n");
    else
      fprintf (stderr, "Taking images using read\n");
    if (pre_capture_command[0])
      fprintf (stderr, "Executing '%s' before each image capture\n",
	       pre_capture_command[0]);
    if (post_capture_command[0])
      fprintf (stderr, "Executing '%s' after each image capture\n",
	       post_capture_command[0]);
    if (backGround)
      fprintf (stderr, "Will switch to run in the background\n");
    else
      fprintf (stderr, "Will not switch to run in the background\n");
  }
  videoIn = (struct vdIn *) calloc (1, sizeof (struct vdIn));
  if (init_videoIn
      (videoIn, (char *) videodevice, width, height, 1, format, grabmethod) < 0)
    exit (1);

  if (cameraControlsEnabled) {
    //Reset all camera controls
    if (verbose >= 1)
      fprintf (stderr, "Resetting camera settings\n");
    v4l2ResetControl (videoIn, V4L2_CID_BRIGHTNESS);
    v4l2ResetControl (videoIn, V4L2_CID_CONTRAST);
    v4l2ResetControl (videoIn, V4L2_CID_SATURATION);
    v4l2ResetControl (videoIn, V4L2_CID_GAIN);

    //Setup Camera Parameters
    if (brightness != 0) {
      if (verbose >= 1)
        fprintf (stderr, "Setting camera brightness to %d\n", brightness);
      v4l2SetControl (videoIn, V4L2_CID_BRIGHTNESS, brightness);
    } else if (verbose >= 1) {
      fprintf (stderr, "Camera brightness level is %d\n", v4l2GetControl (videoIn, V4L2_CID_BRIGHTNESS));
    }
    if (contrast != 0) {
      if (verbose >= 1)
        fprintf (stderr, "Setting camera contrast to %d\n", contrast);
      v4l2SetControl (videoIn, V4L2_CID_CONTRAST, contrast);
    } else if (verbose >= 1) {
      fprintf (stderr, "Camera contrast level is %d\n", v4l2GetControl (videoIn, V4L2_CID_CONTRAST));
    }
    if (saturation != 0) {
      if (verbose >= 1)
        fprintf (stderr, "Setting camera saturation to %d\n", saturation);
      v4l2SetControl (videoIn, V4L2_CID_SATURATION, saturation);
    } else if (verbose >= 1) {
      fprintf (stderr, "Camera saturation level is %d\n", v4l2GetControl (videoIn, V4L2_CID_SATURATION));
    }
    if (gain != 0) {
      if (verbose >= 1)
        fprintf (stderr, "Setting camera gain to %d\n", gain);
      v4l2SetControl (videoIn, V4L2_CID_GAIN, gain);
    } else if (verbose >= 1) {
      fprintf (stderr, "Camera gain level is %d\n", v4l2GetControl (videoIn, V4L2_CID_GAIN));
    }
  }
  ref_time = time (NULL);

  if(backGround)
  {
    if(verbose > 0)
      fprintf (stderr, "starting the daemonising process\n");

    verbose = 0;
    pid_t pid = fork();
    if(pid < 0)
    {
      fprintf (stderr, "Unable to daemonise the process\n");
      exit(EXIT_FAILURE);
    }
    if (pid > 0)
      exit(EXIT_SUCCESS);

    close(STDIN_FILENO);
    close(STDOUT_FILENO);
    close(STDERR_FILENO);
  }

  while (run) {
    if (preDelay > 0) {
      if (verbose > 0)
        fprintf (stderr, "Pre-delay: %d, sleeping...\n", preDelay);

      for (int i = 0; i < preDelay; i++) {
        if (run == 0)
          break;
        // allowing camera to adjust exposure
        if (i == 0 || preDelayCapture) {
          uvcGrab(videoIn);
          videoIn->getPict = 0;
        }

        sleep(1);
      }
    }

    if (pre_capture_command[0]) {
      pre_capture_command[1] = videodevice;

      if (verbose >= 1)
        fprintf(stderr, "Executing '%s %s'\n", pre_capture_command[0], pre_capture_command[1]);
      if (spawn(pre_capture_command, command_wait, verbose)) {
        fprintf(stderr, "Command exited with error\n");
        close_v4l2(videoIn);
        free(videoIn);
        exit(1);
      }
    }

    if (verbose >= 2)
      fprintf (stderr, "Grabbing frame\n");
    if (uvcGrab (videoIn) < 0) {
      fprintf (stderr, "Error grabbing\n");
      close_v4l2 (videoIn);
      free (videoIn);
      exit (1);
    }

    if ((difftime (time (NULL), ref_time) > delay) || delay == 0) {
      if( outputfile[0] != 0 ) {
        save_time = time(NULL);
        timestruct = *(localtime(&save_time));
        strftime(timedfilename, MAXFILENAMELENGTH, outputfile, &timestruct);
        if (verbose >= 1)
	  fprintf (stderr, "Saving image to: %s\n", timedfilename);
        file = fopen (timedfilename, "wb");
        if (file != NULL) {
	  switch (videoIn->formatIn) {
	  case V4L2_PIX_FMT_YUYV:
	    compress_yuyv_to_jpeg (videoIn, file, quality);
	    break;
	  default:
	    imageWithDHT = NULL;
	    verify_jpeg_dht(videoIn->tmpbuffer, videoIn->buf.bytesused, &imageWithDHT, &imageWithDHTSize, verbose);
	    if( imageWithDHT != NULL ) {
	      fwrite (imageWithDHT, imageWithDHTSize, 1, file);
	      free( imageWithDHT );
	    }
	    break;
	  }
	  fclose (file);
	}

        if (post_capture_command[0]) {
          post_capture_command[1] = timedfilename;

	  if (verbose >= 1)
	    fprintf (stderr, "Executing '%s %s'\n", post_capture_command[0],
		   post_capture_command[1]);
	  if (spawn (post_capture_command, command_wait, verbose)) {
	    fprintf (stderr, "Command exited with error\n");
	    close_v4l2 (videoIn);
	    free (videoIn);
	    exit (1);
	  }
        }
      }
      videoIn->getPict = 0;

      ref_time = time (NULL);
    }
    if (delay == 0)
      break;

    if (postDelay > 0) {
      if (verbose > 0)
        fprintf (stderr, "Post-delay: %d, sleeping...\n", postDelay);

      for (int i = 0; i < postDelay; i++) {
        if (run == 0)
          break;
        // allowing camera to adjust exposure
        if (i == 0 || preDelayCapture) {
          uvcGrab(videoIn);
          videoIn->getPict = 0;
        }

        sleep(1);
      }
    }
  }
  close_v4l2 (videoIn);
  free (videoIn);

  return 0;
}
